"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING, Iterable, List, Union

from .exceptions import QueueEmpty, RepeatException
from .tracks import MultiTrack

if TYPE_CHECKING:
    from .player import Player
    from .tracks import Track

__all__ = ("Queue",)


class Queue:
    """
    A class representing a usable Queue.
    """
    def __init__(self, player: Player) -> None:
        self._player: Player = player
        self._tracks: List[Track] = []
        self._currentTrack: int = -1

    def __repr__(self) -> str:
        return f"<Lavapy Queue (Queue={self._tracks})>"

    @property
    def tracks(self) -> List[Track]:
        """Returns a list of track objects."""
        return self._tracks

    @property
    def currentTrack(self) -> int:
        """Returns the current position in the queue."""
        return self._currentTrack

    @property
    def count(self) -> int:
        """Returns the size of the queue."""
        return len(self._tracks)

    @property
    def isEmpty(self) -> bool:
        """Returns whether the queue is empty or not."""
        return self.currentTrack == self.count-1

    def next(self) -> Track:
        """
        Gets the next :class:`Track` in the queue.

        Raises
        ------
        QueueEmpty
            The current queue is empty.
        RepeatException
            Cannot get next track if player is repeating

        Returns
        -------
        Track
            The next track in the queue.
        """
        if self.isEmpty:
            raise QueueEmpty("Queue is empty")
        if self._player.isRepeating:
            raise RepeatException("Cannot get next track if player is repeating")
        self._currentTrack += 1
        return self.tracks[self.currentTrack]

    def previous(self) -> Track:
        """
        Gets the previous :class:`Track` in the queue.

        Raises
        ------
        QueueEmpty
            The current queue is empty.
        RepeatException
            Cannot get previous track if player is repeating

        Returns
        -------
        Track
            The previous track in the queue.
        """
        if self.currentTrack == 0:
            raise QueueEmpty("Queue is empty")
        if self._player.isRepeating:
            raise RepeatException("Cannot get previous track if player is repeating")
        self._currentTrack -= 1
        return self.tracks[self.currentTrack]


    def add(self, track: Track) -> None:
        """
        Adds a :class:`Track` to the queue.

        Parameters
        ----------
        track: Track
            The track to add to the queue.
        """
        self._tracks.append(track)

    def addIterable(self, iterable: Union[MultiTrack, Iterable[Track]]) -> None:
        """
        Adds an iterable to the :class:`Queue`.

        Parameters
        ----------
        iterable: Union[MultiTrack, Iterable[Track]]
            The iterable to add to the queue.
        """
        if isinstance(iterable, MultiTrack):
            iterable = iterable.tracks
        for track in iterable:
            self.add(track)

    def reset(self) -> None:
        """Resets the queue."""
        self.tracks.clear()
        self._currentTrack = 0

    def shuffle(self) -> None:
        """Shuffles the queue but keeps the currently playing track in the same place."""
        tempTrack = self.tracks.pop(self.currentTrack)
        random.shuffle(self.tracks)
        self.tracks.insert(self.currentTrack, tempTrack)



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

from typing import Any, Dict, List, Tuple, Union

from .exceptions import InvalidFilterArgument

__all__ = ("LavapyFilter",
           "Equalizer",
           "Karaoke",
           "Timescale",
           "Tremolo",
           "Vibrato",
           "Rotation",
           "Distortion",
           "ChannelMix",
           "LowPass")


class LavapyFilter:
    """Base Lavapy Filter. Every filter inherit from this."""
    name = ""

    def __init__(self) -> None:
        self._payload: Any = {}

    def __repr__(self) -> str:
        return f"<Lavapy Filter (Payload={self._payload})>"

    @property
    def payload(self) -> Any:
        """Returns the payload to be sent to Lavalink."""
        return self._payload


class Equalizer(LavapyFilter):
    """
    A class representing a usable equalizer.

    Parameters
    ---------
    levels: List[Tuple[int, float]]
        A list of tuple pairs containing a band int and gain float.
    name: str
        A string to name this equalizer.
    """
    name = "equalizer"

    def __init__(self, levels: List[Tuple[int, float]], name: str) -> None:
        super().__init__()
        self._levels: List[Tuple[int, float]] = levels
        self._equalizerName: str = name
        self._payload: List[Dict[str, Union[int, float]]] = self._setup(levels)

    def __repr__(self) -> str:
        return f"<Lavapy Equalizer (Name={self.name}) (Levels={self.levels})>"

    @property
    def levels(self) -> List[Tuple[int, float]]:
        """Returns a list of tuple pairs containing a band int and gain float."""
        return self._levels

    @property
    def equalizerName(self) -> str:
        """Returns the name of this equalizer."""
        return self._equalizerName

    @staticmethod
    def _setup(levels: List[Tuple[int, float]]) -> List[Dict[str, Union[int, float]]]:
        """
        A function to convert self._levels into a dict for sending to Lavalink.

        Parameters
        ----------
        levels: List[Tuple[int, float]]
            A list of tuple pairs containing a band int and gain float.

        Returns
        -------
        List[Dict[str, Union[int, float]]]
            A list of {'band': int, 'gain': float} pairs.
        """
        return [{"band": level[0], "gain": level[1]} for level in levels]

    @classmethod
    def build(cls, levels: List[Tuple[int, float]], name: str = "CustomEqualizer") -> Equalizer:
        """
        Build a custom equalizer with the given levels.

        Parameters
        ----------
        levels: List[Tuple[int, float]]
            A custom list of tuple pairs containing a band int and gain float. You will have to construct this yourself. There should be between 0 and 14 bands with each gain being between -0.25 and 1
        name: str
            An optional string to name this equalizer. If this is not supplied, it will be set to 'CustomEqualizer'.

        Returns
        -------
        Equalizer
            A custom equalizer object.
        """
		    if not (0 <= len(levels) <= 14):
            raise ValueError("There should be between 0 and 14 bands.")
        if not all([-0.25 <= val[1] <= 1 for val in levels]):
            raise ValueError("Each gain should be between -0.25 and 1 bands.")
        return cls(levels, name)

    @classmethod
    def flat(cls) -> Equalizer:
        """
        A flat equalizer. This will not provide a cut or boost to any frequency.

        Returns
        -------
        Equalizer
            A flat equalizer object.
        """
        levels = [(0, 0.0), (1, 0.0), (2, 0.0), (3, 0.0), (4, 0.0),
                  (5, 0.0), (6, 0.0), (7, 0.0), (8, 0.0), (9, 0.0),
                  (10, 0.0), (11, 0.0), (12, 0.0), (13, 0.0), (14, 0.0)]
        return cls(levels, "Flat")


class Karaoke(LavapyFilter):
    """
    Simulates an :class:`Equalizer` which specifically targets removing vocals.

    Attributes
    ----------
    level: float
        The level of the karaoke filter.
    monoLevel: float
        The monolevel of the karaoke filter.
    filterBand: float
        The filter band of the karaoke filter.
    filterWidth: float
        The filter width of the karaoke filter.
    """
    name = "karaoke"

    def __init__(self, level: float = 1.0, monoLevel: float = 1.0, filterBand: float = 220.0, filterWidth: float = 100.0) -> None:
        super().__init__()
        self.level = self._payload["level"] = level
        self.monoLevel = self._payload["monoLevel"] = monoLevel
        self.filterBand = self._payload["filterBand"] = filterBand
        self.filterWidth = self._payload["filterWidth"] = filterWidth

    def __repr__(self) -> str:
        return f"<Lavapy KaraokeFilter (Payload={self._payload})>"


class Timescale(LavapyFilter):
    """
    Changes the speed, pitch and rate of a track. You can make some very cool sound effects with this such as a vaporwave-esque filter which slows the track down a certain amount to produce said effect.

    Attributes
    ----------
    speed: float
        The speed of the timescale filter. This should be more than or equal to 0.
    pitch: float
        The pitch of the timescale filter. This should be more than or equal to 0.
    rate: float
        The rate of the timescale filter. This should be more than or equal to 0.
    """
    name = "timescale"

    def __init__(self, speed: float = 1.0, pitch: float = 1.0, rate: float = 1.0) -> None:
        super().__init__()
        if speed <= 0.0:
            raise InvalidFilterArgument("Speed must be more or equal to than 0.")
        if pitch <= 0.0:
            raise InvalidFilterArgument("Pitch must be more or equal to than 0.")
        if rate <= 0.0:
            raise InvalidFilterArgument("Rate must be more or equal to than 0.")
        self.speed = self._payload["speed"] = speed
        self.pitch = self._payload["pitch"] = pitch
        self.rate = self._payload["rate"] = rate

    def __repr__(self) -> str:
        return f"<Lavapy TimescaleFilter (Payload={self._payload})>"


class Tremolo(LavapyFilter):
    """
    Uses amplification to create a shuddering effect, where the volume quickly oscillates.

    Here is an `Example <https://en.wikipedia.org/wiki/File:Fuse_Electronics_Tremolo_MK-III_Quick_Demo.ogv>`_.

    Attributes
    ----------
    frequency: float
        The frequency of the tremolo filter. This must be bigger than 0.
    depth: float
        The depth of the tremolo filter. This must be between 0 and 1.

    Raises
    ------
    InvalidFilterArgument
        An invalid filter argument has been passed.
    """
    name = "tremolo"

    def __init__(self, frequency: float = 2.0, depth: float = 0.5) -> None:
        super().__init__()
        if frequency < 0.0:
            raise InvalidFilterArgument("Frequency must be more than 0.")
        if depth < 0.0 or depth > 1.0:
            raise InvalidFilterArgument("Depth must be between 0 and 1.")
        self.frequency = self._payload["frequency"] = frequency
        self.depth = self._payload["depth"] = depth

    def __repr__(self) -> str:
        return f"<Lavapy TremoloFilter (Payload={self._payload})>"


class Vibrato(LavapyFilter):
    """
    Similar to :class:`Tremolo` but oscillates the pitch instead of the volume.

    Attributes
    ----------
    frequency: float
        The frequency of the vibrato filter. This must be between 0 and 14.
    depth: float
        The depth of the tremolo filter. This must be between 0 and 1.

    Raises
    ------
    InvalidFilterArgument
        An invalid filter argument has been passed.
    """
    name = "vibrato"

    def __init__(self, frequency: float = 2.0, depth: float = 0.5) -> None:
        super().__init__()
        if frequency < 0.0 or frequency > 14.0:
            raise InvalidFilterArgument("Frequency must be between 0 and 14.")
        if depth < 0.0 or depth > 1.0:
            raise InvalidFilterArgument("Depth must be between 0 and 1.")
        self.frequency = self._payload["frequency"] = frequency
        self.depth = self._payload["depth"] = depth

    def __repr__(self) -> str:
        return f"<Lavapy VibratoFilter (Payload={self._payload})>"


class Rotation(LavapyFilter):
    """
    Rotates the sound around the stereo channels/user headphones aka Audio Panning.

    Here is an `Example <https://en.wikipedia.org/wiki/File:Fuse_Electronics_Tremolo_MK-III_Quick_Demo.ogv>`_ (
    without the reverb).

    Parameters
    ----------
    rotationHz: float
        The frequency of the audio rotating around the listener in hertz (0.2 is similar to the example above).
    """
    name = "rotation"

    def __init__(self, rotationHz: float = 0.0) -> None:
        super().__init__()
        self.rotationHz = self._payload["rotationHz"] = rotationHz

    def __repr__(self) -> str:
        return f"<Lavapy RotationFilter (Payload={self._payload})>"


class Distortion(LavapyFilter):
    """
    Distorts the sound. This can generate some pretty unique audio effects.

    Attributes
    ----------
    sinOffset: float
        The sine offset of the distortion filter.
    sinScale: float
        The sine scale of the distortion filter.
    cosOffset: float
        The cosine offset of the distortion filter.
    cosScale: float
        The cosine scale of the distortion filter.
    tanOffset: float
        The tangent offset of the distortion filter.
    tanScale: float
        The tangent scale of the distortion filter.
    offset: float
        The offset of the distortion filter.
    scale: float
        The scale of the distortion filter.
    """
    name = "distortion"

    def __init__(self, sinOffset: float = 0.0, sinScale: float = 1.0, cosOffset: float = 0.0, cosScale: float = 1.0, tanOffset: float = 0.0, tanScale: float = 1.0, offset: float = 0.0, scale: float = 1.0) -> None:
        super().__init__()
        self.sinOffset = self._payload["sinOffset"] = sinOffset
        self.sinScale = self._payload["sinScale"] = sinScale
        self.cosOffset = self._payload["cosOffset"] = cosOffset
        self.cosScale = self._payload["cosScale"] = cosScale
        self.tanOffset = self._payload["tanOffset"] = tanOffset
        self.tanScale = self._payload["tanScale"] = tanScale
        self.offset = self._payload["offset"] = offset
        self.scale = self._payload["scale"] = scale

    def __repr__(self) -> str:
        return f"<Lavapy DistortionFilter (Payload={self._payload})>"


class ChannelMix(LavapyFilter):
    """
    Mixes both channels (left and right) with a configurable factor on how much each channel affects the other. By
    default, both channel are kept separate from each other. Setting all factors to 0.5 means both channels get the
    same audio.

    Attributes
    ----------
    leftToLeft: float
        The channel mix of left to left. This must be between 0 and 1.
    leftToRight: float
        The channel mix of left to right. This must be between 0 and 1.
    rightToLeft: float
        The channel mix of right to left. This must be between 0 and 1.
    rightToRight: float
        The channel mix of right to right. This must be between 0 and 1.
    """
    name = "channelMix"

    def __init__(self, leftToLeft: float = 1.0, leftToRight: float = 0.0, rightToLeft: float = 0.0, rightToRight: float = 1.0) -> None:
        super().__init__()
        if leftToLeft < 0.0 or leftToLeft > 1.0:
            raise InvalidFilterArgument("LeftToLeft must be between 0 and 1.")
        if leftToRight < 0.0 or leftToRight > 1.0:
            raise InvalidFilterArgument("LeftToRight must be between 0 and 1.")
        if rightToLeft < 0.0 or rightToLeft > 1.0:
            raise InvalidFilterArgument("RightToLeft must be between 0 and 1.")
        if rightToRight < 0.0 or rightToRight > 1.0:
            raise InvalidFilterArgument("RightToRight must be between 0 and 1.")
        self.leftToLeft = self._payload["leftToLeft"] = leftToLeft
        self.leftToRight = self._payload["leftToRight"] = leftToRight
        self.rightToLeft = self._payload["rightToLeft"] = rightToLeft
        self.rightToRight = self._payload["rightToRight"] = rightToRight

    def __repr__(self) -> str:
        return f"<Lavapy ChannelMixFilter (Payload={self._payload})>"


class LowPass(LavapyFilter):
    """
    Suppresses higher frequencies, while allowing lower frequencies to pass through.

    Attributes
    ----------
    smoothing: float
        The smoothing of the low pass filter.
    """
    name = "lowPass"

    def __init__(self, smoothing: float = 20.0) -> None:
        super().__init__()
        self.smoothing = self._payload["smoothing"] = smoothing

    def __repr__(self) -> str:
        return f"<Lavapy LowPassFilter (Payload={self._payload})>"



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

from base64 import b64encode
from typing import Dict, Optional

import aiohttp

from .exceptions import SpotifyAuthException

__all__ = ("SpotifyClient",)


class SpotifyClient:
    """
    Provides an interface for easily initialising and communicating with Spotify.

    Parameters
    ----------
    clientID: str
        The Spotify client ID of the application you want to connect to.
    clientSecret: str
        The Spotify client secret of the application you want to connect to.
    """
    def __init__(self, clientID: str, clientSecret: str) -> None:
        self._clientID: str = clientID
        self._clientSecret: str = clientSecret
        self._session: aiohttp.ClientSession = aiohttp.ClientSession()
        self._accessToken: Optional[str] = None

    def __repr__(self) -> str:
        return f"<Lavapy SpotifyClient)>"

    @property
    def clientID(self) -> str:
        """Returns the client ID of the Spotify client."""
        return self._clientID

    @property
    def clientSecret(self) -> str:
        """Returns the client secret of the Spotify client."""
        return self._clientSecret

    @property
    def session(self) -> aiohttp.ClientSession:
        """Returns the session used for communicating with Spotify."""
        return self._session

    @property
    def accessToken(self) -> Optional[str]:
        """Returns the access token used to authenticate with Spotify."""
        return self._accessToken

    @property
    def authHeaders(self) -> Dict[str, str]:
        """Returns the headers used for authenticating Spotify requests."""
        return {
            "Authorization": f"Bearer {self.accessToken}",
            "Content-Type": "application/json"
        }

    async def _getBearerToken(self) -> None:
        """|coro|

        Gets a Spotify bearer token for use when communicating with Spotify.
        """
        authTokenBytes = f"{self.clientID}:{self.clientSecret}".encode()
        bearerHeaders = {
            "Authorization": f"Basic {b64encode(authTokenBytes).decode()}",
            "Content-Type": "application/x-www-form-urlencoded"
        }
        async with self.session.post("https://accounts.spotify.com/api/token?grant_type=client_credentials", headers=bearerHeaders) as response:
            if response.status != 200:
                raise SpotifyAuthException("An error occurred while authenticating with Spotify.")
            data = await response.json()
        self._accessToken = data["access_token"]



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import datetime
import logging
from typing import TYPE_CHECKING, Any, Dict, Optional, Type, Union

import nextcord.ext

from .exceptions import FilterAlreadyExists, FilterNotApplied, InvalidSeekPosition, RepeatException
from .pool import NodePool
from .queue import Queue
from .tracks import MultiTrack, PartialResource

if TYPE_CHECKING:
    from .filters import LavapyFilter
    from .pool import Node
    from .tracks import Track

__all__ = ("Player",)

logger = logging.getLogger(__name__)


class Player(nextcord.VoiceProtocol):
    """
    Lavapy Player object. This subclasses :class:`nextcord.VoiceProtocol` and such should be treated as one with additions.

    Examples
    --------
    .. code-block:: python

       @commands.command()
       async def connect(self, channel: nextcord.VoiceChannel):
           voice_client = await channel.connect(cls=lavapy.Player)

    .. warning::
        This class should not be created manually but can be subclassed to add additional functionality. You should instead use :meth:`nextcord.VoiceChannel.connect()` and pass the player object to the cls kwarg.

    Parameters
    ----------
    client: Union[:class:`nextcord.Client`, :class:`nextcord.AutoShardedClient`, :class:`nextcord.ext.commands.Bot`, :class:`nextcord.ext.commands.AutoShardedBot`]
        A client or bot object.
    channel: nextcord.VoiceChannel
        A voice channel for the player to connect to.
    """
    def __init__(self, client: Union[nextcord.Client, nextcord.AutoShardedClient, nextcord.ext.commands.Bot, nextcord.ext.commands.AutoShardedBot], channel: nextcord.VoiceChannel) -> None:
        super().__init__(client, channel)
        self.client: Union[nextcord.Client, nextcord.AutoShardedClient, nextcord.ext.commands.Bot, nextcord.ext.commands.AutoShardedBot] = client
        self.channel: nextcord.VoiceChannel = channel
        self._node: Optional[Node] = NodePool.balanced()
        self._track: Optional[Track] = None
        self._volume: float = 1.0
        self._filters: Dict[str, LavapyFilter] = {}
        self._queue: Queue = Queue(self)
        self._voiceState: Dict[str, Any] = {}
        self._connected: bool = False
        self._paused: bool = False
        self._lastUpdateTime: Optional[datetime.datetime] = None
        self._lastPosition: Optional[float] = None
        self._repeat: bool = False

    def __repr__(self) -> str:
        return f"<Lavapy Player (ChannelID={self.channel.id}) (GuildID={self.guild.id})>"

    @property
    def guild(self) -> nextcord.Guild:
        """Returns the guild this player is in."""
        return self.channel.guild

    @property
    def node(self) -> Optional[Node]:
        """Returns the Lavapy node this player is connected to."""
        return self._node

    @property
    def track(self) -> Optional[Track]:
        """Returns the currently playing track if there is one."""
        return self._track

    @property
    def volume(self) -> float:
        """Returns the current volume."""
        return self._volume

    @property
    def filters(self) -> Dict[str, LavapyFilter]:
        """Returns all the currently applied filters if there are any."""
        return self._filters

    @property
    def queue(self) -> Queue:
        """Returns a queue object which can be used to line up and retrieve tracks."""
        return self._queue

    @property
    def position(self) -> float:
        """Returns the current position of the track in seconds. If nothing is playing, this returns 0."""
        if not self.isPlaying:
            return 0

        if self.isPaused:
            return min(self._lastPosition, self.track.length)

        timeSinceLastUpdate = (datetime.datetime.now(datetime.timezone.utc) - self._lastUpdateTime).total_seconds()
        return min(self._lastPosition + timeSinceLastUpdate, self.track.length)

    @property
    def isConnected(self) -> bool:
        """Returns whether the player is connected to a channel or not."""
        return self._connected

    @property
    def isPlaying(self) -> bool:
        """Returns whether the player is currently playing a track or not."""
        return self.isConnected and self.track is not None

    @property
    def isPaused(self) -> bool:
        """Returns whether the player is currently paused or not."""
        return self._paused

    @property
    def isDead(self) -> bool:
        """Returns whether the player is dead or not. A player is considered dead if it has been destroyed and removed from stored players."""
        return self not in self.node.players

    @property
    def isRepeating(self) -> bool:
        """Returns whether the player is repeating the current track or not."""
        return self._repeat

    def _updateState(self, state: Dict[str, Any]) -> None:
        """
        Stores the last update time and the last position.

        Parameters
        ----------
        state: Dict[str, Any]
            The raw info sent by Lavalink.
        """
        # State updates are sent in milliseconds so need to be converted to seconds (/1000)
        logger.debug(f"Received state update for player with guild id {self.guild.id} with data {state}")
        state = state.get("state")
        self._lastUpdateTime = datetime.datetime.fromtimestamp(state.get("time")/1000, datetime.timezone.utc)

        self._lastPosition = state.get("position", 0)/1000

    def _multitrackCheck(self, track: Union[Track, PartialResource, MultiTrack]) -> Track:
        """
        Checks if a resource is a multitrack and if it is then it grabs the first track and pushes the rest to the queue.

        Parameters
        ----------
        track: Union[Track, PartialResource, MultiTrack]
            A Lavapy multitrack resource to check.
        """
        if isinstance(track, MultiTrack):
            temp = track
            track = temp.tracks.pop(0)
            self.queue.addIterable(temp)
        return track

    def startRepeat(self) -> None:
        """
        Repeats the currently playing track.

        Raises
        ------
        RepeatException
            The player is already repeating.
        """
        if self.isRepeating:
            raise RepeatException("The player is already repeating.")
        self._repeat = True

    def stopRepeat(self) -> None:
        """
        Stops repeating the currently playing track.

        Raises
        ------
        RepeatException
            The player is not repeating.
        """
        if not self.isRepeating:
            raise RepeatException("The player is not repeating..")
        self._repeat = False

    async def on_voice_server_update(self, data: Dict[str, str]) -> None:
        """|coro|

        Called when the client connects to a voice channel.

        Parameters
        ----------
        data: Dict[str, str]
            The raw info sent by discord about the voice channel.
        """
        self._voiceState.update({"event": data})
        await self._sendVoiceUpdate()

    async def on_voice_state_update(self, data: Dict[str, Any]) -> None:
        """|coro|

        Called when the client's voice state changes.

        Parameters
        ----------
        data: Dict[str, Any]
            The raw info sent by discord about the client's voice state.
        """
        self._voiceState.update({"sessionId": data["session_id"]})
        channelID = data["channel_id"]
        if channelID is None:
            # Disconnecting
            self._voiceState.clear()
            return
        channel = self.client.get_channel(channelID)
        if channel != self.channel:
            logger.debug(f"Moved player to channel {channel.id}")
            self.channel = channel
        await self._sendVoiceUpdate()

    async def _sendVoiceUpdate(self) -> None:
        """|coro|

        Sends data collected from on_voice_server_update and on_voice_state_update to Lavalink.
        """
        if {"sessionId", "event"} == self._voiceState.keys():
            logger.debug(f"Dispatching voice update for guild {self.guild.id}")
            voiceUpdate = {
                "op": "voiceUpdate",
                "guildId": str(self.guild.id),
                "sessionId": self._voiceState["sessionId"],
                "event": self._voiceState["event"]
            }
            await self.node._send(voiceUpdate)

    async def connect(self, *, timeout: float, reconnect: bool) -> None:
        """|coro|

        Connects the player to a :class:`nextcord.VoiceChannel`.

        Parameters
        ----------
        timeout: float
            The timeout for the connection.
        reconnect: bool
            A bool stating if reconnection is expected.
        """
        await self.guild.change_voice_state(channel=self.channel)
        self.node.players.append(self)
        self._connected = True
        logger.info(f"Connected to voice channel {self.channel.id}")

    async def disconnect(self, *, force: bool = False) -> None:
        """|coro|

        Disconnects the player from a :class:`nextcord.VoiceChannel`.

        Parameters
        ----------
        force: bool
            Whether to force the disconnection. This is currently not used.
        """
        await self.guild.change_voice_state(channel=None)
        self._connected = False
        self.cleanup()
        logger.info(f"Disconnected from voice channel {self.channel.id}")

    async def play(self, track: Union[Track, PartialResource, MultiTrack], startTime: int = 0, endTime: int = 0, volume: float = 1, replace: bool = True, pause: bool = False) -> Optional[Track]:
        """|coro|

        Plays a given resource. If this resource is a :class:`Track`, it is played normally. However, if it is a
        :class:`PartialResource`, then it is searched for and played (if the result of the search is a list of
        tracks, then the first is played). If the resource is a :class:`MultiTrack` then the first track is played
        and the rest and sent to the :class:`Queue`.

        Parameters
        ----------
        track: Union[Track, PartialResource, MultiTrack]
            The Lavapy resource to play.
        startTime: int
            The position in milliseconds to start at. By default, this is the beginning.
        endTime: int
            The position in milliseconds to end at. By default, this is the end.
        volume: int
            The volume at which the player should play the track at. This should be a value between 0 and 5.
        replace: bool
            A bool stating if the current track should be replaced or not.
        pause: bool
            A bool stating if the track should start paused.

        Returns
        -------
        Optional[Track]
            The currently playing Lavapy track.
        """
        if self.isPlaying and not replace:
            return
        track = self._multitrackCheck(track)
        if isinstance(track, PartialResource):
            track = await self.node.getTracks(track.cls, track.query)
            if track is None:
                return
            if isinstance(track, list):
                track = track[0]
            track = self._multitrackCheck(track)
		if not (0 <= volume <= 5):
            raise ValueError("Volume must be a value between 0 and 5.")
        newTrack = {
            "op": "play",
            "guildId": str(self.guild.id),
            "track": track.id,
            "startTime": str(startTime),
            "volume": str(volume),
            "noReplace": not replace,
            "pause": pause
        }
        if endTime > 0:
            newTrack["endTime"] = str(endTime)
        self._track = track
        self._volume = volume
        await self.node._send(newTrack)
        if self.queue.currentTrack == -1:
            # First time a new song is playing
            self.queue.tracks.insert(0, track)
            self.queue._currentTrack += 1
        logger.debug(f"Started playing track {self.track.title} in {self.channel.id}")
        return track

    async def stop(self) -> None:
        """|coro|

        Stops the currently playing :class:`Track`.
        """
        stop = {
            "op": "stop",
            "guildId": str(self.guild.id)
        }
        tempTrack = self.track
        self._track = None
        await self.node._send(stop)
        logger.debug(f"Stopped playing track {tempTrack.title} in {self.channel.id}")

    async def pause(self) -> None:
        """|coro|

        Pauses the :class:`Player` if it was playing.
        """
        await self._togglePause(True)
        logger.debug(f"Paused track in {self.channel.id}")

    async def resume(self) -> None:
        """|coro|

        Resumes the :class:`Player` if it was paused.
        """
        await self._togglePause(False)
        logger.debug(f"Resumed track in {self.channel.id}")

    async def _togglePause(self, pause: bool) -> None:
        """|coro|

        Toggles the player's pause state.

        Parameters
        ----------
        pause: bool
            A bool stating whether the player's paused state should be True or False.
        """
        pause = {
            "op": "pause",
            "guildId": str(self.guild.id),
            "pause": str(pause)
        }
        self._paused = pause
        await self.node._send(pause)

    async def seek(self, position: int) -> None:
        """|coro|

        Seek to a given position.

        Parameters
        ----------
        position: int
            The position to seek to.

        Raises
        ------
        InvalidIdentifier
            Seek position is bigger than the track length.
        """
        if position > self.track.length:
            raise InvalidSeekPosition("Seek position is bigger than track length.")
        seek = {
            "op": "seek",
            "guildId": str(self.guild.id),
            "position": position
        }
        await self.node._send(seek)
        logger.debug(f"Seeked to position {position} for guild {self.guild.id}")

    async def setVolume(self, volume: int) -> None:
        """|coro|

        Changes the :class:`Player`'s volume.

        Parameters
        ----------
        volume: int
            The new volume. This should be a value between 0 and 5 where 1 is 100%.
        """
        if not (0 <= volume <= 5):
            raise ValueError("Volume must be a value between 0 and 5.")
		self._volume = max(min(volume, 5), 0)
        volume = {
            "op": "volume",
            "guildId": str(self.guild.id),
            "volume": str(self.volume)
        }
        await self.node._send(volume)
        logger.debug(f"Set volume to {volume} for guild {self.guild.id}")

    async def moveTo(self, channel: nextcord.VoiceChannel) -> None:
        """|coro|

        Moves the player to another :class:`nextcord.VoiceChannel`. If this is None, then the player will disconnect.

        Parameters
        ----------
        channel: nextcord.VoiceChannel
            The voice channel to move to.
        """
        await self.guild.change_voice_state(channel=channel)

    async def addFilter(self, filter: LavapyFilter) -> None:
        """|coro|

        Adds a :class:`LavapyFilter` to the player.

        .. warning::
            This requires Lavalink 3.4 or greater..

        Parameters
        ----------
        filter: LavapyFilter
            The filter to apply to the player.

        Raises
        ------
        FilterAlreadyExists
            The specific filter is already applied.
        """
        name = filter.name
        if filter.name in self.filters.keys():
            raise FilterAlreadyExists(f"Filter <{name}> is already applied. Please remove it first.")
        self.filters[name] = filter
        await self._updateFilters()
        logger.debug(f"Added filter {name} with payload {filter.payload} for guild {self.guild.id}")

    async def removeFilter(self, filter: Union[LavapyFilter, Type[LavapyFilter]]) -> None:
        """|coro|

        Removes a :class:`LavapyFilter` based on its name.

        .. warning::
            This requires Lavalink 3.4 or greater.

        Parameters
        ----------
        filter: Union[LavapyFilter, Type[LavapyFilter]]
            The filter to remove. This can either be a non-initialised class like `lavapy.Equalizer` or an initialised one like `lavapy.Equalizer.flat()`.

        Raises
        ------
        FilterNotApplied
            The specific filter is not applied.
        """
        name = filter.name
        if name not in self.filters.keys():
            raise FilterNotApplied(f"{name} is not applied.")
        del self.filters[name]
        await self._updateFilters()
        logger.debug(f"Removed filter {name} for guild {self.guild.id}")

    async def _updateFilters(self) -> None:
        """|coro|

        Collects all applied filters and sends them to Lavalink.
        """
        filterPayload = {
            "op": "filters",
            "guildId": str(self.guild.id),
            "volume": self.volume/100,
        }
        for key, value in self.filters.items():
            filterPayload[value.name] = value.payload
        await self.node._send(filterPayload)

    async def resetFilter(self, filter: Type[LavapyFilter]) -> None:
        """|coro|

        Resets a :class:`LavapyFilter`.

        .. warning::
            This requires Lavalink 3.4 or greater.

        Parameters
        ----------
        filter: Type[LavapyFilter]
            The filter to reset. This should be a non-initialised class like `lavapy.Equalizer` or `lavapy.Karaoke`.

        Raises
        ------
        FilterNotApplied
            The specific filter is not applied.
        """
        name = filter.name
        if name not in self.filters.keys():
            raise FilterNotApplied(f"{name} is not applied.")
        if name == "equalizer":
            # noinspection PyUnresolvedReferences
            self.filters[name] = filter.flat()
        else:
            self.filters[name] = filter()
        await self._updateFilters()

    async def destroy(self) -> None:
        """|coro|

        Disconnects and destroys the player.
        """
        logger.debug(f"Destroying player with guild id {self.guild.id}")
        await self.disconnect()
        self.node.players.remove(self)
        destroyPayload = {
            "op": "destroy",
            "guildId": str(self.guild.id)
        }
        await self.node._send(destroyPayload)



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Dict, Optional

import aiohttp

from .backoff import ExponentialBackoff
from .events import LavapyEvent, TrackEndEvent, TrackExceptionEvent, TrackStartEvent, TrackStuckEvent, WebsocketClosedEvent, WebsocketOpenEvent
from .stats import Stats
from .tracks import Track

if TYPE_CHECKING:
    from .player import Player
    from .pool import Node

logger = logging.getLogger(__name__)

__all__ = ("Websocket",)


class Websocket:
    """
    Lavapy Websocket object.

    Parameters
    ----------
    node: Node
        The Lavapy node object which manages this websocket.
    """
    def __init__(self, node: Node) -> None:
        self._node: Node = node
        self._connection: Optional[aiohttp.client.ClientWebSocketResponse] = None
        self._listener: Optional[asyncio.Task] = None
        asyncio.create_task(self.connect())

    def __repr__(self) -> str:
        return f"<Lavapy Websocket (Domain={self.node.host}:{self.node.port}) (Connected={self.connected})>"

    @property
    def node(self) -> Node:
        """Returns the Lavapy node which manages this websocket."""
        return self._node

    @property
    def connected(self) -> bool:
        """Returns whether the websocket is connected or not."""
        return self._connection is not None and not self.connection.closed

    @property
    def connection(self) -> Optional[aiohttp.client.ClientWebSocketResponse]:
        """Returns the websocket connection."""
        return self._connection

    @property
    def listener(self) -> Optional[asyncio.Task]:
        """Returns the task which pings the Lavalink server for updates."""
        return self._listener

    def getPlayer(self, guildID: int) -> Player:
        """
        Returns a Lavapy :class:`Player` object based on a specific :class:`nextcord.Guild` ID.

        Parameters
        ----------
        guildID: int
            The guild ID to get a Lavapy player object for.

        Returns
        -------
        Player
            A Lavapy player object.
        """
        return [] or [player for player in self.node.players if player.guild.id == guildID][0]

    async def connect(self) -> None:
        """|coro|

        Establishes the connection to the Lavalink server.

        Raises
        ------
        aiohttp.client.ClientConnectorError
            The domain name is invalid.
        """
        headers = {
            "Authorization": self.node.password,
            "User-Id": str(self.node.client.user.id),
            "Client-Name": "Lavapy"
        }
        logger.debug(f"Attempting connection with for node {self.node.identifier}")
        try:
            self._connection = await self.node.session.ws_connect(self.node.websocketUri, headers=headers, heartbeat=self.node.heartbeat)
        except Exception as error:
            if isinstance(error, aiohttp.WSServerHandshakeError) and error.status == 401:
                logger.error(f"Authorisation failed for node {self.node.identifier}")
            else:
                logger.error(f"Connection failure for node {self.node.identifier} with error {error}")
            return
        self._listener = self.node.client.loop.create_task(self.createListener())
        logger.debug(f"Connection established for node {self.node.identifier}")
        event = WebsocketOpenEvent(self.node)
        await self.dispatchEvent(f"lavapy_{event.event}", event.payload)

    async def disconnect(self) -> None:
        """|coro|

        Closes the connection to the Lavalink server.
        """
        logger.debug(f"Closing connection for node {self.node.identifier}")
        self.listener.cancel()
        await self.connection.close()

    async def createListener(self) -> None:
        """|coro|

        Continuously pings the Lavalink server for updates and events.
        """
        backoff = ExponentialBackoff()
        while True:
            msg = await self.connection.receive()
            if msg.type is aiohttp.WSMsgType.CLOSED:
                logger.debug(f"Websocket closed for node {self.node.identifier} with info {msg.extra}")
                retry = backoff.delay()
                logger.debug(f"Retrying connection in {retry} seconds")
                await asyncio.sleep(retry)
            else:
                if msg.data == 1011:
                    logger.error(f"Internal Lavalink error encountered with node {self.node.identifier}. Terminating without retries. Consider updating your Lavalink server.")
                    self.listener.cancel()
                asyncio.create_task(self.processListener(msg.json()))

    async def processListener(self, data: Dict[str, Any]) -> None:
        """|coro|

        Processes data received from the Lavalink server gathered in :meth:`createListener()`.

        Parameters
        ----------
        data: Dict[str, Any]
            The raw data received from Lavalink.
        """
        op = data.get("op")
        if op == "playerUpdate":
            try:
                player = self.getPlayer(int(data["guildId"]))
                player._updateState(data)
            except IndexError:
                # Player has recently sent a destroy op so don't update state
                pass
        elif op == "event":
            event = await self.getEventPayload(data["type"], data)
            await self.dispatchEvent(f"lavapy_{event.event}", event.payload)
        elif op == "stats":
            self.node._stats = Stats(self.node, data)

    async def getEventPayload(self, name: str, data: Dict[str, Any]) -> LavapyEvent:
        """|coro|

        Processes events received from Lavalink sent from :meth:`processListener()`.

        Parameters
        ----------
        name: str
            The event name.
        data: Dict[str, Any]
            The event payload sent by Lavalink.

        Returns
        -------
        LavapyEvent
            The LavapyEvent object which corresponds to the event payload.
        """
        logger.debug(f"Received event payload {data}")
        if name == "WebSocketClosedEvent":
            return WebsocketClosedEvent(self.node, data)

        player = self.getPlayer(int(data["guildId"]))
        track = await self.node.buildTrack(Track, data["track"])
        if name == "TrackStartEvent":
            return TrackStartEvent(player, track)
        elif name == "TrackEndEvent":
            if player.isRepeating:
                await player.play(player.track)
            elif data["reason"] == "FINISHED":
                player._track = None
            return TrackEndEvent(player, track, data)
        elif name == "TrackExceptionEvent":
            return TrackExceptionEvent(player, track, data)
        elif name == "TrackStuckEvent":
            return TrackStuckEvent(player, track, data)

    async def dispatchEvent(self, event: str, payload: Dict[str, Any]) -> None:
        """|coro|

        Dispatches events to Discord.

        Parameters
        ----------
        event: str
            The event name.
        payload: Dict[str, Any]
            The payload to dispatch with the event. This is sent to `**kwargs`.
        """
        logger.debug(f"Dispatching event {event} with payload {payload}")
        self.node.client.dispatch(event, **payload)



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional

if TYPE_CHECKING:
    from .player import Player
    from .pool import Node
    from .tracks import Track

__all__ = ("LavapyEvent",
           "TrackStartEvent",
           "TrackEndEvent",
           "TrackExceptionEvent",
           "TrackStuckEvent",
           "WebsocketOpenEvent",
           "WebsocketClosedEvent")


class LavapyEvent:
    """
    Base Lavapy event. Every event inherits from this.

    If you want to listen to these events, use a :meth:`nextcord.ext.commands.Bot.listen()`.

    Parameters
    ----------
    event: str
        The event name.
    player: Optional[Player]
        A Lavapy player object.
    """
    def __init__(self, event: str, player: Optional[Player]) -> None:
        self._event: str = event
        self._payload: Dict[str, Any] = {}
        if player is not None:
            self._payload["player"] = player

    def __repr__(self) -> str:
        return f"<Lavapy LavapyEvent (Payload={self.payload})>"

    @property
    def event(self) -> str:
        """Returns the event name."""
        return self._event

    @property
    def payload(self) -> Dict[str, Any]:
        """Returns a dict containing the payload sent to nextcord. This must be parsed to `**kwargs`."""
        return self._payload


class TrackStartEvent(LavapyEvent):
    """
    Fired when a :class:`Track` starts playing. This can be listened to with:

    .. code-block:: python

        @bot.listen()
        async def on_lavapy_track_start(player, track):
            ...

    Parameters
    ----------
    player: Player
        A Lavapy player object.
    track: Track
        A Lavapy track object.
    """
    def __init__(self, player: Player, track: Track) -> None:
        super().__init__("track_start", player)
        self._payload["track"] = track

    def __repr__(self) -> str:
        return f"<Lavapy TrackStartEvent (Payload={self.payload})>"


class TrackEndEvent(LavapyEvent):
    """
    Fired when a :class:`Track` stops playing. This can be listened to with:

    .. code-block:: python

        @bot.listen()
        async def on_lavapy_track_end(player, track, reason):
            ...

    Parameters
    ----------
    player: Player
        A Lavapy player object.
    track: Track
        A Lavapy track object.
    data: Dict[str, Any]
        The raw event data.
    """
    def __init__(self, player: Player, track: Track, data: Dict[str, Any]) -> None:
        super().__init__("track_end", player)
        self._payload["track"] = track
        self._payload["reason"] = data["reason"]

    def __repr__(self) -> str:
        return f"<Lavapy TrackStopEvent (Payload={self.payload})>"


class TrackExceptionEvent(LavapyEvent):
    """
    Fired when a :class:`Track` error has occurred in Lavalink. This can be listened to with:

    .. code-block:: python

        @bot.listen()
        async def on_lavapy_track_exception(player, track, exception):
            ...

    Parameters
    ----------
    player: Player
        A Lavapy player object.
    track: Track
        A Lavapy track object.
    data: Dict[str, Any]
        The raw event data.
    """
    def __init__(self, player: Player, track: Track, data: Dict[str, Any]) -> None:
        super().__init__("track_exception", player)
        self._payload["track"] = track
        if data.get("error"):
            # User is running Lavalink <= 3.3
            self._payload["exception"] = data["error"]
        else:
            # User is running Lavalink >= 3.4
            self._payload["exception"] = data["exception"]

    def __repr__(self) -> str:
        return f"<Lavapy TrackExceptionEvent (Payload={self.payload})>"


class TrackStuckEvent(LavapyEvent):
    """
    Fired when a :class:`Track` is stuck and cannot be played. This can be listened to with:

    .. code-block:: python

        @bot.listen()
        async def on_lavapy_track_stuck(player, track, threshold):
            pass

    Parameters
    ----------
    player: Player
        A Lavapy player object.
    track: Track
        A Lavapy track object.
    data: Dict[str, Any]
        The raw event data.
    """
    def __init__(self, player: Player, track: Track, data: Dict[str, Any]) -> None:
        super().__init__("track_stuck", player)
        self._payload["track"] = track
        self._payload["threshold"] = data["thresholdMs"]

    def __repr__(self) -> str:
        return f"<Lavapy TrackStuckEvent (Payload={self.payload})>"


class WebsocketOpenEvent(LavapyEvent):
    """
    Fired when a websocket connection to a :class:`Node` is open. This can be listened to with:

    .. code-block:: python

        @bot.listen()
        async def on_lavapy_websocket_open(node):
            pass

    Parameters
    ----------
    node: lavapy.pool.Node
        A Lavapy node object.
    """
    def __init__(self, node: Node) -> None:
        super().__init__("websocket_open", None)
        self._payload["node"] = node

    def __repr__(self) -> str:
        return f"<Lavapy WebsocketOpenEvent (Payload={self.payload})>"


class WebsocketClosedEvent(LavapyEvent):
    """
    Fired when a websocket connection to a :class:`Node` is closed. This can be listened to with:

    .. code-block:: python

        @bot.listen()
        async def on_lavapy_websocket_closed(node, reason, code, byRemote):
            pass

    Parameters
    ----------
    node: lavapy.pool.Node
        A Lavapy node object.
    data: Dict[str, Any]
        The raw event data.
    """
    def __init__(self, node: Node, data: Dict[str, Any]) -> None:
        super().__init__("websocket_closed", None)
        self._payload["node"] = node
        self._payload["reason"] = data["reason"]
        self._payload["code"] = data["code"]
        self._payload["byRemote"] = data["byRemote"]

    def __repr__(self) -> str:
        return f"<Lavapy WebsocketClosedEvent (Payload={self.payload})>"



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple, Type, Union

from lavapy.tracks import MultiTrack, Playable, Track

if TYPE_CHECKING:
    from lavapy.pool import Node

__all__ = ("decodeSpotifyQuery",
           "SpotifyBase",
           "SpotifyTrack",
           "SpotifyPlaylist",
           "SpotifyAlbum")


def decodeSpotifyQuery(query: str) -> Type[Playable]:
    """
    Decodes a query into a :class:`Playable` type which can be searched.

    Parameters
    ----------
    query: str
        The query to decode.

    Returns
    -------
    Type[Playable]
        The playable type which can be searched.
    """
    if re.compile("https://open\.spotify\.com/track").match(query):
        return SpotifyTrack
    elif re.compile("https://open\.spotify\.com/playlist").match(query):
        return SpotifyPlaylist
    elif re.compile("https://open\.spotify\.com/track/album").match(query):
        return SpotifyAlbum
    else:
        return SpotifyTrack


async def spotifyGetDetails(cls: Type[SpotifyBase], query: str, node: Node) -> Union[str, List[str]]:
    """|coro|

    Gets details about a Spotify track.

    Parameters
    ----------
    cls: Type[SpotifyPlayable]
        The Lavapy Spotify resource to get the search type for.
    query: str
        The query to search Spotify for.
    node: Node
        The Lavapy node to use for querying the Spotify web API.

    Returns
    -------
    Union[str, List[str]]
        A query string or list of query strings.
    """
    regexResult = re.compile("https://open\.spotify\.com/.+/(?P<identifier>.+)\?").match(query)
    if cls._spotifyType == "track":
        if not regexResult:
            return f"ytsearch:{query}"
        else:
            identifier = regexResult.group("identifier")
            async with node.spotifyClient.session.get(f"https://api.spotify.com/v1/tracks/{identifier}", headers=node.spotifyClient.authHeaders) as response:
                data = await response.json()
            return f'ytsearch:{data["artists"][0]["name"]} - {data["name"]}'
    elif cls._spotifyType == "playlist":
        url = f"https://api.spotify.com/v1/playlists/{regexResult.group('identifier')}/tracks"
    elif cls._spotifyType == "album":
        url = f"https://api.spotify.com/v1/albums/{regexResult.group('identifier')}/tracks"
    # noinspection PyUnboundLocalVariable
    async with node.spotifyClient.session.get(url, headers=node.spotifyClient.authHeaders) as response:
        data = await response.json()
    trackArr: List[Dict[str, Any]] = data["items"]
    nextUrl = data["next"]
    while nextUrl:
        async with node.spotifyClient.session.get(nextUrl, headers=node.spotifyClient.authHeaders) as response:
            data = await response.json()
        trackArr.extend(data["items"])
        nextUrl = data["next"]
    return [f'ytsearch:{track["track"]["artists"][0]["name"]} - {track["track"]["name"]}' for track in trackArr]


async def spotifyGetMultitrackName(cls: Type[SpotifyBase], query: str, node: Node) -> str:
    """|coro|

    Parameters
    ----------
    cls: Type[SpotifyPlayable]
        The Lavapy Spotify resource to create an instance of.
    query: str
        The query to search Spotify for.
    node: Node
        The Lavapy node to use for querying the Spotify web API.

    Returns
    -------
    str
        The multitrack's name.
    """
    regexResult = re.compile("https://open\.spotify\.com/.+/(?P<identifier>.+)\?").match(query)
    if cls._spotifyType == "track":
        return ""
    elif cls._spotifyType == "playlist":
        async with node.spotifyClient.session.get(f"https://api.spotify.com/v1/playlists/{regexResult.group('identifier')}", headers=node.spotifyClient.authHeaders) as response:
            data = await response.json()
    elif cls._spotifyType == "album":
        async with node.spotifyClient.session.get(f"https://api.spotify.com/v1/albums/{regexResult.group('identifier')}", headers=node.spotifyClient.authHeaders) as response:
            data = await response.json()
    return data["name"]


class SpotifyBase(Playable):
    """A class which abstracts away a couple functions used for all Spotify resources."""
    _spotifyType: str
    _queryGetter: Callable = spotifyGetDetails
    _getMultitrackName: Callable = spotifyGetMultitrackName


class SpotifyTrack(Track, SpotifyBase):
    """A track created using a search to Spotify."""
    _spotifyType: str = "track"

    def __repr__(self) -> str:
        return f"<Lavapy SpotifyTrack (YoutubeIdentifier={self.identifier})>"


class SpotifyPlaylist(MultiTrack, SpotifyBase):
    """A playlist created using a search to Spotify."""
    _spotifyType: str = "playlist"

    def __repr__(self) -> str:
        return f"<Lavapy SpotifyPlaylist (Name={self.name}) (Track count={len(self.tracks)})>"


class SpotifyAlbum(MultiTrack, SpotifyBase):
    """An album created using a search to Spotify."""
    _spotifyType: str = "album"
    _trackCls: Type[Track] = SpotifyTrack

    def __repr__(self) -> str:
        return f"<Lavapy SpotifyAlbum (Name={self.name}) (Track count={len(self.tracks)})>"



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import logging
import random
import string
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union

import aiohttp

from .exceptions import BuildTrackError, InvalidNodeSearch, LavalinkException, LoadTrackError, NodeOccupied, NoNodesConnected, WebsocketAlreadyExists
from .ext.spotify.tracks import SpotifyBase
from .stats import Stats
from .websocket import Websocket

if TYPE_CHECKING:
    import nextcord.ext
    from nextcord import VoiceRegion
    from .ext.spotify.client import SpotifyClient
    from .player import Player
    from .tracks import MultiTrack, Playable, Track

__all__ = ("NodePool",
           "Node",)

logger = logging.getLogger(__name__)


class NodePool:
    """
    Lavapy NodePool class. This holds all the :class:`Node` objects created with :meth:`createNode()`.
    """
    _nodes: Dict[str, Node] = {}

    @classmethod
    def nodes(cls) -> Dict[str, Node]:
        """Returns a mapping of identifier to node objects."""
        return cls._nodes

    @classmethod
    async def createNode(cls, *, client: Union[nextcord.Client, nextcord.AutoShardedClient, nextcord.ext.commands.Bot, nextcord.ext.commands.AutoShardedBot], host: str, port: int, password: str, region: Optional[VoiceRegion] = None, secure: bool = False, heartbeat: int = 60, spotifyClient: Optional[SpotifyClient] = None, identifier: Optional[str] = None, ssl: Optional[bool] = None) -> Node:
        """|coro|

        Creates a Lavapy :class:`Node` object and stores it for later use.

        Parameters
        ----------
        client: Union[:class:`nextcord.Client`, :class:`nextcord.AutoShardedClient`, :class:`nextcord.ext.commands.Bot`, :class:`nextcord.ext.commands.AutoShardedBot`]
            The client or bot object to assign to this node.
        host: str
            The IP address of the Lavalink server.
        port: int
            The port of the Lavalink server.
        password: str
            The password to the Lavalink server.
        region: Optional[:class:`nextcord.VoiceRegion`]
            The voice region to assign to this node.
        secure: bool
            Whether to connect securely to the Lavalink server or not.
        heartbeat: int
            How many seconds to wait before sending a ping message and waiting for a response from the Lavalink server.
        spotifyClient: Optional[SpotifyClient]
            A Lavapy Spotify client for interacting with Spotify.
        identifier: Optional[str]
            The unique identifier for this node. If not supplied, it will be generated for you.
		ssl: Optional[bool]
			SSL validation mode. None for default SSL check, False for skip SSL certificate validation.

        Raises
        ------
        NodeOccupied
            A node with the given identifier already exists.

        Returns
        -------
        Node
            A Lavapy node object.
        """
        if identifier is None:
            identifier = "".join(random.choices(string.ascii_letters + string.digits, k=8))

        if identifier in cls._nodes:
            raise NodeOccupied(f"A node with the identifier <{identifier}> already exists.")

        node = Node(client, host, port, password, region, secure, heartbeat, spotifyClient, identifier, ssl)
        cls._nodes[identifier] = node
        await node.connect()
        await node._initialiseExtensions()
        return node

    @classmethod
    def minPlayers(cls) -> Node:
        """
        An algorithm which selects the best Lavapy :class:`Node` based on the amount of players.

        Raises
        ------
        NoNodesConnected
            There are currently no nodes connected with the provided options.
        InvalidNodeSearch
            No nodes could be found with the current algorithm.

        Returns
        -------
        Node
            A Lavapy node object.
        """
        if not cls._nodes:
            raise NoNodesConnected("There are currently no nodes connected.")
        result = sorted(cls._nodes.values(), key=lambda x: len(x.players))
        try:
            return result[0]
        except KeyError:
            raise InvalidNodeSearch("No nodes could be found using the minPlayers algorithm.")

    @classmethod
    def balanced(cls) -> Node:
        """
        An algorithm which selects the best Lavapy :class:`Node` based on each node's penalty.

        Raises
        ------
        NoNodesConnected
            There are currently no nodes connected with the provided options.
        InvalidNodeSearch
            No nodes could be found with the current algorithm.

        Returns
        -------
        Node
            A Lavapy node object.
        """
        if not cls._nodes:
            raise NoNodesConnected("There are currently no nodes connected.")
        result = sorted(cls._nodes.values(), key=lambda x: x.penalty)
        try:
            return result[0]
        except KeyError:
            raise InvalidNodeSearch("No nodes could be found using the balanced algorithm.")

    @classmethod
    def identifier(cls, identifier: str):
        """
        An algorithm which selects the best Lavapy :class:`Node` based on its identifier.

        Raises
        ------
        NoNodesConnected
            There are currently no nodes connected with the provided options.
        InvalidNodeSearch
            No nodes could be found with the current algorithm.
        """
        if not cls._nodes:
            raise NoNodesConnected("There are currently no nodes connected.")
        try:
            return cls._nodes[identifier]
        except KeyError:
            raise InvalidNodeSearch(f"No nodes could be found with identifier {identifier}.")

    @classmethod
    def closestNode(cls, region: VoiceRegion):
        """
        An algorithm which selects the best Lavapy :class:`Node` based on how close it is to the desired target.

        Raises
        ------
        NoNodesConnected
            There are currently no nodes connected with the provided options.
        InvalidNodeSearch
            No nodes could be found with the current algorithm.
        """
        if not cls._nodes:
            raise NoNodesConnected("There are currently no nodes connected.")
        result = [node for node in cls._nodes.values() if node.region == region]
        try:
            return result[0]
        except KeyError:
            raise InvalidNodeSearch(f"No nodes could be found with region {region}.")

    @classmethod
    def extension(cls, extension: Type[Playable]):
        """
        An algorithm which selects the best Lavapy :class:`Node` based on the amount of players.

        Raises
        ------
        NoNodesConnected
            There are currently no nodes connected with the provided options.
        InvalidNodeSearch
            No nodes could be found with the current algorithm.
        """
        if not cls._nodes:
            raise NoNodesConnected("There are currently no nodes connected.")
        if issubclass(extension, SpotifyBase):
            result = [node for node in cls._nodes.values() if node.spotifyClient is not None]
        else:
            result = list(cls._nodes.values())
        try:
            return result[0]
        except KeyError:
            raise InvalidNodeSearch(f"No nodes could be found with extension {extension}.")


class Node:
    """
    Lavapy Node object.

    .. warning::
        This class should not be created manually. Please use :meth:`NodePool.createNode()` instead.
    """
    def __init__(self, client: Union[nextcord.Client, nextcord.AutoShardedClient, nextcord.ext.commands.Bot, nextcord.ext.commands.AutoShardedBot], host: str, port: int, password: str, region: Optional[nextcord.VoiceRegion], secure: bool, heartbeat: int, spotifyClient: Optional[SpotifyClient], identifier: str, ssl: bool) -> None:
        self._client: Union[nextcord.Client, nextcord.AutoShardedClient, nextcord.ext.commands.Bot, nextcord.ext.commands.AutoShardedBot] = client
        self._host: str = host
        self._port: int = port
        self._password: str = password
        self._region: Optional[nextcord.VoiceRegion] = region
        self._secure: bool = secure
        self._heartbeat: int = heartbeat
        self._spotifyClient: Optional[SpotifyClient] = spotifyClient
        self._identifier: str = identifier
        self._players: List[Player] = []
        self._stats: Optional[Stats] = None
        self._session: aiohttp.ClientSession = aiohttp.ClientSession()
        self._websocket: Optional[Websocket] = None
        self._websocketUri: str = f"{'wss' if self._secure else 'ws'}://{self.host}:{self.port}"
        self._restUri: str = f"{'https' if self._secure else 'http'}://{self.host}:{self.port}"
        self._ssl: bool = ssl

    def __repr__(self) -> str:
        return f"<Lavapy Node (Domain={self.host}:{self.port}) (Identifier={self.identifier}) (Region={self.region}) (Players={len(self.players)})>"

    @property
    def client(self) -> Union[nextcord.Client, nextcord.AutoShardedClient, nextcord.ext.commands.Bot, nextcord.ext.commands.AutoShardedBot]:
        """Returns the client or bot object assigned to this node."""
        return self._client

    @property
    def host(self) -> str:
        """Returns the IP address of the Lavalink server."""
        return self._host

    @property
    def port(self) -> int:
        """Returns the port of the Lavalink server."""
        return self._port

    @property
    def password(self) -> str:
        """Returns the password to the Lavalink server."""
        return self._password

    @property
    def region(self) -> Optional[nextcord.VoiceRegion]:
        """Returns the voice region assigned to this node."""
        return self._region

    @property
    def secure(self) -> bool:
        """Returns whether the connection to the Lavalink server is secure or not."""
        return self._secure

    @property
    def heartbeat(self) -> int:
        """Returns the amount of seconds between ping requests."""
        return self._heartbeat

    @property
    def spotifyClient(self) -> Optional[SpotifyClient]:
        """Returns the Spotify client associated with this node if there is one."""
        return self._spotifyClient

    @property
    def identifier(self) -> str:
        """Returns the unique identifier for this node."""
        return self._identifier

    @property
    def players(self) -> List[Player]:
        """Returns a list of all Lavapy player objects which are connected to this node."""
        return self._players

    @property
    def stats(self) -> Optional[Stats]:
        """Returns useful information sent by Lavalink about this node."""
        return self._stats

    @property
    def session(self) -> aiohttp.client.ClientSession:
        """Returns the session used for sending and getting data."""
        return self._session

    @property
    def websocketUri(self) -> str:
        """Returns the URI used for initialising the websocket."""
        return self._websocketUri

    @property
    def restUri(self) -> str:
        """Returns the URI used for communicating with the Lavalink server."""
        return self._restUri

    @property
    def penalty(self) -> float:
        """Returns the load balancing penalty for this node."""
        if self.stats is None:
            return 0.0
        return self.stats.penalty.total

    @property
    def ssl(self):
        """Returns the SSL validation mode."""
        return self._ssl

    async def _initialiseExtensions(self) -> None:
        """|coro|

        Initialises all the extensions linked to this :class:`Node`.
        """
        if self.spotifyClient is not None:
            logger.debug(f"Initialising spotify extension for node {self.identifier}")
            await self.spotifyClient._getBearerToken()

    async def connect(self) -> None:
        """|coro|

        Initialises the websocket connection to the Lavalink server.

        Raises
        ------
        WebsocketAlreadyExists
            The websocket for this node already exists.
        """
        logger.debug(f"Connecting to the Lavalink server at {self.host}:{self.port}")
        if self._websocket is None:
            self._websocket = Websocket(self)
        else:
            raise WebsocketAlreadyExists("Websocket already initialised.")

    async def disconnect(self, *, force: bool = False) -> None:
        """|coro|

        Disconnects this :class:`Node` and removes it from the :class:`NodePool`.

        Parameters
        ----------
        force: bool
            Whether to force the disconnection. This is currently not used.
        """
        logger.debug(f"Disconnecting node {self.identifier}")
        for player in self.players:
            await player.disconnect(force=force)

        try:
            await self._websocket.disconnect()
            del NodePool._nodes[self.identifier]
        except Exception as e:
            logger.debug(f"Failed to remove node {self.identifier} with error {e}")

    async def _getData(self, endpoint: str, params: Dict[str, str]) -> Tuple[Dict[str, Any], aiohttp.ClientResponse]:
        """|coro|

        Make a request to Lavalink with a given endpoint and return a response.

        Parameters
        ----------
        endpoint: str
            The endpoint to query from Lavalink.
        params: Dict[str, str]
            A dict containing additional info to send to Lavalink.

        Returns
        -------
        Tuple[Dict[str, Any], :class:`aiohttp.ClientResponse`]
            A tuple containing the response from Lavalink as well as the websocket response to determine the status of the request.
        """
        logger.debug(f"Getting endpoint {endpoint} with data {params}")
        headers = {
            "Authorization": self.password
        }
        async with await self.session.get(f"{self.restUri}/{endpoint}", headers=headers, params=params, ssl=self._ssl) as req:
            data = await req.json()
        return data, req

    async def _send(self, payload: Dict[str, Any]) -> None:
        """|coro|

        Send a payload to Lavalink without a response.

        Parameters
        ----------
        payload: Dict[str, Any]
            The payload to send to Lavalink.
        """
        logger.debug(f"Sending payload {payload}")
        await self._websocket.connection.send_json(payload)

    async def buildTrack(self, cls: Type[Track], id: str) -> Track:
        """|coro|

        Builds a :class:`Track` from a base64 track ID.

        Parameters
        ----------
        cls: Type[Track]
            The Lavapy resource to return an instance of
        id: str
            The base64 track ID.

        Raises
        ------
        BuildTrackError
            An error occurred while building the track.

        Returns
        -------
        Track
            A Lavapy track object.
        """
        logger.debug(f"Building track with cls {cls} and id {id}")
        track, response = await self._getData("decodetrack", {"track": id})
        if response.status != 200:
            raise BuildTrackError("A error occurred while building the track.", track)
        return cls(id, track)

    async def getTracks(self, cls: Type[Playable], query: str) -> Optional[Union[Track, List[Track], MultiTrack]]:
        """|coro|

        Gets data about a :class:`Track` or :class:`MultiTrack` from Lavalink.

        Parameters
        ----------
        cls: Type[Playable]
            The Lavapy resource to create an instance of.
        query: str
            The query to search for via Lavalink.

        Returns
        -------
        Optional[Union[Track, List[Track], MultiTrack]]
            A Lavapy resource which can be used to play music.
        """
        logger.info(f"Getting data with cls {cls} and query: {query}")
        data, response = await self._getData("loadtracks", {"identifier": query})
        if response.status != 200:
            raise LavalinkException("Invalid response from lavalink.")

        loadType = data.get("loadType")
        if loadType == "LOAD_FAILED":
            raise LoadTrackError(f"Track failed to load with data: {data}.")
        elif loadType == "NO_MATCHES":
            return None
        elif loadType == "TRACK_LOADED":
            trackInfo = data["tracks"][0]
            # noinspection PyTypeChecker
            return cls(trackInfo["track"], trackInfo["info"])
        elif loadType == "SEARCH_RESULT":
            # noinspection PyTypeChecker
            return [cls(element["track"], element["info"]) for element in data["tracks"]]
        elif loadType == "PLAYLIST_LOADED":
            playlistInfo = data["playlistInfo"]
            # noinspection PyTypeChecker
            return cls(playlistInfo["name"], [cls._trackCls(track["track"], track["info"]) for track in data["tracks"]])



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
__all__ = ("LavapyException",
           "NoNodesConnected",
           "NodeOccupied",
           "InvalidSeekPosition",
           "WebsocketAlreadyExists",
           "QueueEmpty",
           "InvalidFilterArgument",
           "FilterAlreadyExists",
           "FilterNotApplied",
           "InvalidNodeSearch",
           "RepeatException",
           "LavalinkException",
           "LoadTrackError",
           "BuildTrackError")


class LavapyException(Exception):
    """Base Lavapy exception. Every exception inherits from this."""


class NoNodesConnected(LavapyException):
    """Exception raised when an operation is attempted with nodes and there are none connected."""


class NodeOccupied(LavapyException):
    """Exception raised when node identifiers conflict."""


class InvalidSeekPosition(LavapyException):
    """Exception raised when an invalid seek position is passed."""


class WebsocketAlreadyExists(LavapyException):
    """Exception raised when a new websocket connection is attempted but it already exists."""


class QueueEmpty(LavapyException):
    """Exception raised when attempting to get a track from an empty queue."""


class InvalidFilterArgument(LavapyException):
    """Exception raised when an invalid argument is passed to a filter."""


class FilterAlreadyExists(LavapyException):
    """Exception raised when a new filter is attempted to be applied, but it already exists."""


class FilterNotApplied(LavapyException):
    """Exception raised when a filter is attempted to be removed, but it is not applied."""


class InvalidNodeSearch(LavapyException):
    """Exception raised when a search for a node is invalid."""


class RepeatException(LavapyException):
    """Exception raised when an error occurred when enabling or disabling repeat."""


class LavalinkException(LavapyException):
    """Base exception raised when an error occurs communicating with Lavalink."""


class LoadTrackError(LavalinkException):
    """Exception raised when an error occurred when loading a track."""


class BuildTrackError(LavalinkException):
    """Exception raised when an error occurred when building a track."""



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import random

__all__ = ("ExponentialBackoff",)


class ExponentialBackoff:
    """
    An implementation of the exponential backoff algorithm. This provides a convenient interface to implement an
    exponential backoff for reconnecting or retrying transmissions in a distributed network. Once instantiated,
    the delay method will return the next interval to wait for when retrying a connection or transmission.

    Parameters
    ----------
    base: int
        The base delay in seconds. Changing this changes how fast the delay increases.
    maxRetries: int
        The maximum amount of retries allowed. Changing this changes how big the delay can get.
    """
    def __init__(self, base: int = 1, maxRetries: int = 20) -> None:
        self._base: int = base
        self._maxRetries: int = maxRetries

        self._retries: int = 0

        rand = random.Random()
        rand.seed()

        self._rand = rand.uniform

    def __repr__(self) -> str:
        return f"<Lavapy ExponentialBackoff (Base={self.base}) (Retries={self.retries}) (MaxRetries={self.maxRetries})>"

    @property
    def base(self) -> int:
        """Returns the base on the equation."""
        return self._base

    @property
    def retries(self) -> int:
        """Returns the amount of current retries."""
        return self._retries

    @property
    def maxRetries(self) -> int:
        """Returns the max amount of retries."""
        return self._maxRetries

    def delay(self) -> float:
        """
        Computes the next delay. This is a value between 0 and (base*2)^(retries+1) where if the amount of retries
        currently done is bigger than maxRetries, then retries resets.
        """
        self._retries += 1

        if self._retries > self._maxRetries:
            self._retries = 1

        return self._rand(0, self._base * 2 ** self._retries)


import os
import re
from setuptools import setup

with open("lavapy/__init__.py") as init:
    version = re.search(r'__version__ = "([\'0-9\'].[\'0-9\'].[\'0-9\'](?:[ba][\'0-9\'])*)"', init.read()).group(1)

with open("README.rst") as rdme:
    readme = rdme.read()

with open("requirements.txt") as rqmnts:
    requirements = rqmnts.read().splitlines()

if os.getenv("READTHEDOCS") == "True":
    with open("docs/requirementsRTD.txt") as rqmnts:
        requirements.extend(rqmnts.read().splitlines())

setup(
    name="nextlava",
    version=version,
    description="A powerful and robust Python library built from the ground up for interacting with Lavalink.",
    long_description=readme,
    author="Aspect1103 & rexwu1104",
    author_email="jack.ashwell1@gmail.com",
    license="MIT",
    url="https://github.com/Aspect1103/Lavapy",
    packages=["lavapy", "lavapy.ext.spotify"],
    keywords=["lavapy", "lavalink", "nextocrd"],
    python_requires=">=3.8",
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed"
    ]
)



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

if TYPE_CHECKING:
    from .pool import Node

__all__ = ("Stats",
           "Penalty")


class Stats:
    """
    Stores useful information sent by Lavalink about this :class:`Node`.

    Attributes
    ----------
    node: Node
        The Lavapy node these stats are about.
    uptime: int
        How long this node has been up for.
    players: int
        How many players are connected to this node.
    playingPlayers: int
        How many players that are connected to this node are playing tracks.
    memoryReservable: int
        How much memory is reserved by the system.
    memoryUsed: int
        How much memory is used by the system.
    memoryFree: int
        How much memory is free in the system.
    memoryAllocated: int
        How much memory is allocated to Lavalink.
    cpuCores: int
        How many cpu cores the system has.
    systemLoad: float
        The load on the system as a percentage.
    lavalinkLoad: float
        The load on the Lavalink server as a percentage.
    framesSent: int
        How many frames have been sent to Lavalink by this node.
    framesDeficit: int
        How many frames have been lost that were sent by this node.
    framesNulled: int
        How many frames have errored that were sent by this node.
    """
    def __init__(self, node: Node, data: Dict[str, Any]) -> None:
        self.node: Node = node
        self.uptime: int = data["uptime"]
        self.players: int = data["players"]
        self.playingPlayers: int = data["playingPlayers"]

        memory: Dict[str, Any] = data["memory"]
        self.memoryReservable: int = memory["reservable"]
        self.memoryUsed: int = memory["used"]
        self.memoryFree: int = memory["free"]
        self.memoryAllocated: int = memory["allocated"]

        cpu: Dict[str, Any] = data["cpu"]
        self.cpuCores: int = cpu["cores"]
        self.systemLoad: float = cpu["systemLoad"]
        self.lavalinkLoad: float = cpu["lavalinkLoad"]

        frameStats: Dict[str, Any] = data.get("frameStats", {})
        self.framesSent: int = frameStats.get("sent", -1)
        self.framesDeficit: int = frameStats.get("deficit", -1)
        self.framesNulled: int = frameStats.get("nulled", -1)

        self.penalty: Penalty = Penalty(self)

    def __repr__(self) -> str:
        return f"<Lavapy Stats (Node={self.node})>"


class Penalty:
    """Represents a load balancing penalty for use when assigning a Lavapy :class:`Node` object."""
    def __init__(self, stats: Stats):
        self.playerPenalty: int = stats.playingPlayers
        self.cpuPenalty: float = 1.05 ** (100 * stats.systemLoad) * 10 - 10
        self.nullFramePenalty: float = 0
        self.deficitFramePenalty: float = 0

        if stats.framesNulled != -1:
            self.nullFramePenalty = (1.03 ** (500 * (stats.framesNulled / 3000))) * 300 - 300
            self.nullFramePenalty *= 2

        if stats.framesDeficit != -1:
            self.deficitFramePenalty = (
                1.03 ** (500 * (stats.framesDeficit / 3000))
            ) * 600 - 600

        self.total: float = (self.playerPenalty + self.cpuPenalty + self.nullFramePenalty + self.deficitFramePenalty)



"""
MIT License

Copyright (c) 2021-present Aspect1103

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Type, Union

import lavapy

if TYPE_CHECKING:
    from .pool import Node

__all__ = ("decodeQuery",
           "Playable",
           "PartialResource",
           "Track",
           "MultiTrack",
           "YoutubeTrack",
           "YoutubeMusicTrack",
           "SoundcloudTrack",
           "LocalTrack",
           "YoutubePlaylist")


def decodeQuery(query: str) -> Type[Playable]:
    """
    Decodes a query into a :class:`Playable` type which can be searched.

    Parameters
    ----------
    query: str
        The query to decode.

    Returns
    -------
    Type[Playable]
        The playable type which can be searched.
    """
    if re.compile("https://www.youtube.com/watch\?v=.+").match(query):
        return YoutubeTrack
    elif re.compile("https://music.youtube.com/watch\?v=.+").match(query):
        return YoutubeMusicTrack
    elif re.compile("https://soundcloud.com/(?!discover).+").match(query) and "sets" not in query:
        return SoundcloudTrack
    elif re.compile("https://cdn.discordapp.com/.+.mp3").match(query):
        return LocalTrack
    elif re.compile("https://www.youtube.com/playlist\?list=.+").match(query):
        return YoutubePlaylist
    else:
        return YoutubeTrack


async def defaultQueryGetter(cls: Type[Playable], query: str, _) -> str:
    """|coro|

    Forms the query for a specific :class:`Playable`.

    Parameters
    ----------
    cls: Type[Playable]
        The self to get the search type for.
    query: str
        The query to modify with a search type.

    Returns
    -------
    str
        The modified query.
    """
    if not re.compile("https://.+/").match(query):
        return f"{cls._searchType}:{query}"
    return query


class Playable:
    """
    The base class for all Lavapy resources. This supports both query searches and identifier searches.

    .. warning::
        This class should not be created manually. Instead use a subclass of :class:`Track` or :class:`MultiTrack`.
    """
    _searchType: str
    _trackCls: Optional[Type[Track]]
    _queryGetter: Callable = defaultQueryGetter
    _getMultitrackName: Optional[Callable] = None

    @classmethod
    async def search(cls, query: str, node: Node = None, returnFirst: bool = True, partial: bool = False) -> Optional[Union[Playable, Track, List[Track], PartialResource, List[PartialResource], MultiTrack]]:
        """|coro|

        Performs a search to Lavalink for a specific resource.

        Parameters
        ----------
        query: str
            The query to search for.
        node: Node
            The Lavapy Node to use for searching. If this is not supplied, a random one from the node pool will be retrieved.
        returnFirst: bool
            Whether to return only the first result or not.
        partial: bool
            Whether to return a Lavapy partial resource or not.

        Returns
        -------
        Optional[Union[Playable, Track, List[Track], PartialResource, List[PartialResource], MultiTrack]]
            A Lavapy resource or a list of resources which can be used to play music.
        """
        if node is None:
            node = lavapy.NodePool.extension(cls)
        newQuery = await cls._queryGetter(cls, query, node)
        if partial:
            if isinstance(newQuery, list):
                # This will only run with extensions
                multitrackName = await cls._getMultitrackName(cls, query, node)
                return cls(multitrackName, [PartialResource(YoutubeTrack, temp) for temp in newQuery])
            return PartialResource(cls, newQuery)
        if isinstance(newQuery, str):
            tracks = await node.getTracks(cls, newQuery)
        else:
            temp = []
            for i in newQuery:
                result = await node.getTracks(YoutubeTrack, i)
                temp.append(result[0])
            multitrackName = await cls._getMultitrackName(cls, query, node)
            tracks = cls(multitrackName, temp)
        if tracks is not None:
            if isinstance(tracks, list) and returnFirst:
                return tracks[0]
            return tracks

    def __init__(self, *data: Any) -> None:
        """This is just here to stop :meth:`Node.getTracks()` being upset about unexpected arguments."""


class PartialResource:
    """
    A class which searches for the given query at playtime.

    .. warning::
        It is advised not to create this manually, however, it is possible to do so.

    Parameters
    ----------
    cls: Type[Playable]
        The resource to create a instance of at playtime.
    query: str
        The query to search for at playtime.

    .. warning::
        This object will only search for the given query at playtime. Full resource information will be missing until it has been searched. It is advised not to create this manually, however, it is possible to do so.
    """
    def __init__(self, cls: Type[Playable], query: str) -> None:
        self._cls = cls
        self._query: str = query

    def __repr__(self) -> str:
        return f"<Lavapy PartialResource (Cls={self.cls}) (Query={self.query})>"

    @property
    def cls(self) -> Type[Playable]:
        """Returns the resource which will be created at playtime."""
        return self._cls

    @property
    def query(self) -> str:
        """Returns the query which will be searched for at playtime."""
        return self._query


class Track:
    """
    The base class for all Lavapy Track objects.

    .. warning::
        It is advised not to create this manually, however, it is possible to do so.

    Parameters
    ----------
    id: str
        The unique base64 track ID which can be used to rebuild a track.
    info: Dict[str, Any]
        The raw track info.
    """
    def __init__(self, id: str, info: Dict[str, Any]) -> None:
        self._id: str = id
        self._info: Dict[str, Any] = info
        self._identifier: str = info["identifier"]
        self._isSeekable: bool = info["isSeekable"]
        self._author: str = info["author"]
        self._length: int = info["length"]
        self._isStream: bool = info["isStream"]
        self._type: str = info["sourceName"]
        self._title: str = info["title"]
        self._uri: str = info["uri"]

    def __repr__(self) -> str:
        return f"<Lavapy Track (Identifier={self.identifier}) (Type={self.type})>"

    @property
    def id(self) -> str:
        """Returns the base64 track ID."""
        return self._id

    @property
    def identifier(self) -> str:
        """Returns the track's unique identifier."""
        return self._identifier

    @property
    def isSeekable(self) -> bool:
        """Returns whether a track is seekable or not."""
        return self._isSeekable

    @property
    def author(self) -> str:
        """Returns the author of the track."""
        return self._author

    @property
    def length(self) -> int:
        """Returns the duration of the track in seconds."""
        return self._length

    @property
    def isStream(self) -> bool:
        """Returns whether the track is a stream or not."""
        return self._isStream

    @property
    def type(self) -> str:
        """Returns the source site of the track."""
        return self._type

    @property
    def title(self) -> str:
        """Returns the title of the track."""
        return self._title

    @property
    def uri(self) -> str:
        """Returns the URI of the track."""
        return self._uri


class MultiTrack:
    """
    The base class for all Lavapy MultiTrack resources. These could be playlists or albums.

    .. warning::
        It is advised not to create this manually, however, it is possible to do so.

    Parameters
    ----------
    name: str
        The name of the playlist.
    tracks: List[Track]
        The playlist's tracks as a list of Lavapy Track objects.
    """
    def __init__(self, name: str, tracks: List[Track]) -> None:
        self._name: str = name
        self._tracks: List[Track] = tracks

    def __repr__(self) -> str:
        return f"<Lavapy MultiTrack (Name={self.name}) (Track count={len(self.tracks)})>"

    @property
    def name(self) -> str:
        """Returns the name of the playlist."""
        return self._name

    @property
    def tracks(self) -> List[Track]:
        """Returns the playlist's tracks."""
        return self._tracks


class YoutubeTrack(Track, Playable):
    """A track created using a search to Youtube."""
    _searchType: str = "ytsearch"

    def __repr__(self) -> str:
        return f"<Lavapy YoutubeTrack (Identifier={self.identifier})>"

    @property
    def thumbnail(self) -> str:
        """Returns the URI to the thumbnail of this track."""
        return f"https://img.youtube.com/vi/{self.identifier}/maxresdefault.jpg"


class YoutubeMusicTrack(Track, Playable):
    """A track created using a search to Youtube Music."""
    _searchType: str = "ytmsearch"

    def __repr__(self) -> str:
        return f"<Lavapy YoutubeMusicTrack (Identifier={self.identifier})>"


class SoundcloudTrack(Track, Playable):
    """A track created using a search to Soundcloud."""
    _searchType: str = "scsearch"

    def __repr__(self) -> str:
        return f"<Lavapy SoundcloudTrack (Identifier={self.identifier})>"


class LocalTrack(Track, Playable):
    """A track created using a search to a Discord mp3 file."""

    def __repr__(self) -> str:
        return f"<Lavapy LocalTrack (Identifier={self.identifier})>"


class YoutubePlaylist(MultiTrack, Playable):
    """A playlist created using a search to Youtube."""
    _trackCls: Type[Track] = YoutubeTrack

    def __repr__(self) -> str:
        return f"<Lavapy YoutubePlaylist (Name={self.name}) (Track count={len(self.tracks)})>"
